/**
 * \file src/input/mzidentml/mzidentmlreader.cpp
 * \date 24/11/2022
 * \author Olivier Langella
 * \brief new method to read mzIdentML XML files
 */


/*******************************************************************************
 * Copyright (c) 2022 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of i2MassChroQ.
 *
 *     i2MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     i2MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with i2MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#include "mzidentmlreader.h"
#include <pappsomspp/psm/deepprot/deepprotenum.h>


#include <pappsomspp/exception/exceptionnotfound.h>
#include <pappsomspp/exception/exceptioninterrupted.h>
#include "../../utils/utils.h"
#include <QDebug>


QString
MzIdentMlReader::CvParam::toString() const
{
  return QString("%1 %2 %3 %4").arg(cvRef).arg(accession).arg(name).arg(value);
}


QString
MzIdentMlReader::UserParam::toString() const
{
  return QString("%1 %2").arg(name).arg(value);
}


MzIdentMlReader::MzIdentMlReader(
  pappso::UiMonitorInterface *p_monitor,
  Project *p_project,
  IdentificationGroup *p_identification_group,
  IdentificationMzIdentMlFile *p_identification_data_source)
{
  mp_monitor                  = p_monitor;
  mp_project                  = p_project;
  mp_identificationGroup      = p_identification_group;
  mp_identificationDataSource = p_identification_data_source;
  msp_msrun                   = p_identification_data_source->getMsRunSp();
}


MzIdentMlReader::~MzIdentMlReader()
{
}


IdentificationEngine
MzIdentMlReader::getIdentificationEngine() const
{
  if(m_IdentificationEngineMap.size() == 1)
    {
      return m_IdentificationEngineMap.begin()->second;
    }
  if(m_IdentificationEngineMap.size() == 0)
    {
      return IdentificationEngine::unknown;
    }
  else
    {
      throw pappso::PappsoException(
        "Unable to read MzIdentML file containing multiple identification "
        "engines results");
    }
}


void
MzIdentMlReader::readStream()
{
  // mp_monitor->setStatus("reading X!Tandem result file");
  if(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "MzIdentML")
        {
          while(m_qxmlStreamReader.readNextStartElement())
            {
              // cvList
              if(m_qxmlStreamReader.name().toString() == "cvList")
                {
                  m_qxmlStreamReader.skipCurrentElement();
                }
              // AnalysisSoftwareList
              else if(m_qxmlStreamReader.name().toString() ==
                      "AnalysisSoftwareList")
                {
                  while(readAnalysisSoftware())
                    {
                    }
                }
              // SequenceCollection
              else if(m_qxmlStreamReader.name().toString() ==
                      "SequenceCollection")
                {
                  while(readSequenceCollectionItem())
                    {
                    }
                }
              // AnalysisCollection
              else if(m_qxmlStreamReader.name().toString() ==
                      "AnalysisCollection")
                {
                  while(readAnalysisCollectionItem())
                    {
                    }
                }
              // AnalysisProtocolCollection
              else if(m_qxmlStreamReader.name().toString() ==
                      "AnalysisProtocolCollection")
                {
                  m_qxmlStreamReader.skipCurrentElement();
                }

              // DataCollection
              else if(m_qxmlStreamReader.name().toString() == "DataCollection")
                {
                  while(readDataCollectionItem())
                    {
                    }
                }
              else
                {
                  m_qxmlStreamReader.raiseError(
                    QObject::tr("element %1 not implemented")
                      .arg(m_qxmlStreamReader.name()));
                  m_qxmlStreamReader.skipCurrentElement();
                }
            }
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("Not an MzIdentML input file"));
          m_qxmlStreamReader.skipCurrentElement();
        }
    }
}

bool
MzIdentMlReader::readAnalysisSoftware()
{


  /** @brief stores the current analysis software id
   */
  QString analysisSoftwareId;


  /** @brief tells if the software name has been found and is handled by the
   * parser
   */
  IdentificationEngine analysisSotwareNameFound = IdentificationEngine::unknown;
  if(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "AnalysisSoftware")
        {

          // <AnalysisSoftware version="0.0.9" name="DeepProt" id="as1">


          analysisSoftwareId =
            m_qxmlStreamReader.attributes().value("id").toString();
          QString software_name =
            m_qxmlStreamReader.attributes().value("name").toString();

          if(software_name == "DeepProt")
            {
              analysisSotwareNameFound = IdentificationEngine::DeepProt;

              m_qxmlStreamReader.skipCurrentElement();
            }
          else
            {
              while(m_qxmlStreamReader.readNextStartElement())
                {
                  if(m_qxmlStreamReader.name().toString() == "SoftwareName")
                    {
                      while(m_qxmlStreamReader.readNextStartElement())
                        {
                          if(m_qxmlStreamReader.name().toString() == "cvParam")
                            {
                              CvParam cv_param = readCvParam();
                              if(cv_param.accession == "MS:1001476")
                                {
                                  analysisSotwareNameFound =
                                    IdentificationEngine::XTandem;
                                }
                              else if(cv_param.accession == "MS:1002048")
                                {
                                  analysisSotwareNameFound =
                                    IdentificationEngine::MSGFplus;
                                }
                            }
                        }
                    }
                }
            }


          switch(analysisSotwareNameFound)
            {
              case IdentificationEngine::DeepProt:
                break;

              case IdentificationEngine::MSGFplus:
                break;
              case IdentificationEngine::XTandem:
                break;

              default:
                m_qxmlStreamReader.raiseError(
                  QObject::tr(
                    "identification results from %1 are not supported yet, "
                    "Please contact "
                    "the PAPPSO team.")
                    .arg(software_name));
                return false;
            }


          m_IdentificationEngineMap.insert(
            std::pair<QString, IdentificationEngine>(analysisSoftwareId,
                                                     analysisSotwareNameFound));
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("Not an MzIdentML input file"));
          m_qxmlStreamReader.skipCurrentElement();
          return false;
        }
      return true;
    }
  return false;
}

bool
MzIdentMlReader::readSequenceCollectionItem()
{
  if(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "DBSequence")
        {
          readDBSequence();
        }
      else if(m_qxmlStreamReader.name().toString() == "Peptide")
        {
          readPeptide();
        }
      // PeptideEvidence
      else if(m_qxmlStreamReader.name().toString() == "PeptideEvidence")
        {
          readPeptideEvidence();
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in MzIdentML input file, %1 no DBSequence")
              .arg(m_qxmlStreamReader.name()));
          m_qxmlStreamReader.skipCurrentElement();
          return false;
        }
      return true;
    }
  return false;
}

void
MzIdentMlReader::readDBSequence()
{

  // attributes.value("base_name")
  // ProteinXtpSp sp_xtp_protein = _current_protein.makeProteinXtpSp();
  ProteinXtpSp protein = ProteinXtp().makeProteinXtpSp();
  protein.get()->setAccession(
    m_qxmlStreamReader.attributes().value("accession").toString());

  protein = mp_project->getProteinStore().getInstance(protein);


  FastaFileSp fastaFile = mp_project->getFastaFileStore().getInstance(FastaFile(
    m_qxmlStreamReader.attributes().value("searchDatabase_ref").toString()));

  mp_identificationDataSource->addFastaFile(fastaFile);
  m_FastaFileIdMap.insert(std::pair<QString, FastaFileSp>(
    m_qxmlStreamReader.attributes().value("searchDatabase_ref").toString(),
    fastaFile));


  m_ProteinXmlIdMap.insert(std::pair<QString, ProteinXtpSp>(
    m_qxmlStreamReader.attributes().value("id").toString(), protein));

  while(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "cvParam")
        {
          CvParam cv_param = readCvParam();

          if(cv_param.accession == "MS:1001088")
            {
              // protein description
              protein.get()->setDescription(cv_param.value);
            }
          else if(cv_param.accession == "MS:1001195")
            {
              // PSI-MS MS:1001195 decoy DB type reverse
              protein.get()->setIsDecoy(true);
            }
          else
            {

              m_qxmlStreamReader.raiseError(
                QObject::tr("unknown cvParam for DBSequence tag : %1")
                  .arg(cv_param.toString()));
            }
        }
      else if(m_qxmlStreamReader.name().toString() == "Seq")
        {

          protein.get()->setSequence(m_qxmlStreamReader.readElementText());
        }
      else
        {

          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in MzIdentML/DBSequence unexpected %1 tag")
              .arg(m_qxmlStreamReader.name()));
          m_qxmlStreamReader.skipCurrentElement();
        }
    }

  // searchDatabase_ref="SearchDB_1"

  protein.get()->setFastaFileP(fastaFile.get());
}

MzIdentMlReader::CvParam
MzIdentMlReader::readCvParam()
{
  CvParam cv_param;
  cv_param.cvRef = m_qxmlStreamReader.attributes().value("cvRef").toString();
  cv_param.accession =
    m_qxmlStreamReader.attributes().value("accession").toString();
  cv_param.name  = m_qxmlStreamReader.attributes().value("name").toString();
  cv_param.value = m_qxmlStreamReader.attributes().value("value").toString();
  cv_param.unitAccession =
    m_qxmlStreamReader.attributes().value("unitAccession").toString();
  cv_param.unitName =
    m_qxmlStreamReader.attributes().value("unitName").toString();
  cv_param.unitCvRef =
    m_qxmlStreamReader.attributes().value("unitCvRef").toString();
  m_qxmlStreamReader.skipCurrentElement();
  return cv_param;
}

MzIdentMlReader::UserParam
MzIdentMlReader::readUserParam()
{
  UserParam user_param;

  user_param.name  = m_qxmlStreamReader.attributes().value("name").toString();
  user_param.value = m_qxmlStreamReader.attributes().value("value").toString();
  m_qxmlStreamReader.skipCurrentElement();

  return user_param;
}


void
MzIdentMlReader::readPeptide()
{
  PeptideXtpSp peptide;

  QString xml_id = m_qxmlStreamReader.attributes().value("id").toString();

  // PeptideSequence
  if(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "PeptideSequence")
        {
          peptide =
            PeptideXtp(m_qxmlStreamReader.readElementText().simplified())
              .makePeptideXtpSp();
        }
      else
        {

          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in MzIdentML/Peptide no PeptideSequence"));
        }
    }

  while(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "Modification")
        {
          //      <Modification monoisotopicMassDelta="-0.02682025649"
          //      location="1">
          //          <cvParam accession="-0.0268203" cvRef="PSI-MOD" name=""/>
          //      </Modification>

          Modification modification;
          modification.monoisotopicMassDelta = m_qxmlStreamReader.attributes()
                                                 .value("monoisotopicMassDelta")
                                                 .toDouble();
          modification.location =
            m_qxmlStreamReader.attributes().value("location").toUInt();
          while(m_qxmlStreamReader.readNextStartElement())
            {
              if(m_qxmlStreamReader.name().toString() == "cvParam")
                {
                  modification.cvParam = readCvParam();


                  // qDebug() << "startElement_aa ";
                  pappso::AaModificationP modif = nullptr;


                  qDebug() << modification.monoisotopicMassDelta;
                  if(modification.cvParam.accession != "")
                    {
                      qDebug() << modification.cvParam.accession;
                      if(modification.cvParam.accession.startsWith("UNIMOD:"))
                        {
                          modif = Utils::translateAaModificationFromUnimod(
                            modification.cvParam.accession);
                        }
                      else
                        {
                          // hope it is psi mod:
                          if(modification.cvParam.accession.startsWith("MOD:"))
                            {
                              modif = pappso::AaModification::getInstance(
                                modification.cvParam.accession);
                            }
                          else if(modification.cvParam.accession ==
                                  "MS:1001460")
                            {
                              //[Term]
                              // id: MS:1001460
                              // name: unknown modification
                              // def: "This term should be given if the
                              // modification was unknown." [PSI:PI] is_a:
                              // MS:1001471 ! peptide modification details
                            }
                          else
                            {
                              qInfo() << "MzIdentMlSaxHandler::endElement_"
                                         "Modification unknown "
                                         "modification "
                                      << modification.cvParam.accession << " "
                                      << modification.cvParam.name;
                            }
                        }
                    }

                  if(modif == nullptr)
                    {
                      modif =
                        Utils::guessAaModificationPbyMonoisotopicMassDelta(
                          modification.monoisotopicMassDelta);
                    }

                  if(modification.location == 0)
                    {
                      peptide.get()->addAaModification(modif, 0);
                    }
                  else
                    {
                      peptide.get()->addAaModification(
                        modif, modification.location - 1);
                    }
                }
              else
                {

                  m_qxmlStreamReader.raiseError(
                    QObject::tr("Error in MzIdentML/Peptide/Modification "
                                "unexpected %1 tag")
                      .arg(m_qxmlStreamReader.name()));
                }
            }
        }
      else
        {

          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in MzIdentML/Peptide unexpected %1 tag")
              .arg(m_qxmlStreamReader.name()));
        }
    }

  peptide = mp_project->getPeptideStore().getInstance(peptide);


  m_PeptideIdMap.insert(std::pair<QString, PeptideXtpSp>(xml_id, peptide));
}

bool
MzIdentMlReader::readPeptideEvidence()
{
  qDebug();
  MzidPeptideEvidence pe;

  auto itprot = m_ProteinXmlIdMap.find(
    m_qxmlStreamReader.attributes().value("dBSequence_ref").toString());
  if(itprot == m_ProteinXmlIdMap.end())
    {
      m_qxmlStreamReader.raiseError(QObject::tr("dBSequence_ref %1 not defined")
                                      .arg(m_qxmlStreamReader.attributes()
                                             .value("dBSequence_ref")
                                             .toString()));
      return false;
    }
  qDebug();
  pe.protein = itprot->second;


  auto itpep = m_PeptideIdMap.find(
    m_qxmlStreamReader.attributes().value("peptide_ref").toString());
  if(itpep == m_PeptideIdMap.end())
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("peptide_ref %1 not defined")
          .arg(
            m_qxmlStreamReader.attributes().value("peptide_ref").toString()));
      return false;
    }
  qDebug();
  pe.peptide = itpep->second;

  pe.start   = m_qxmlStreamReader.attributes().value("start").toUInt() - 1;
  pe.end     = m_qxmlStreamReader.attributes().value("end").toUInt() - 1;
  pe.isDecoy = false;
  if(m_qxmlStreamReader.attributes().value("isDecoy").toString() == "true")
    {
      pe.isDecoy = true;
    }
  qDebug();
  m_MzidPeptideEvidenceIdMap.insert(std::pair<QString, MzidPeptideEvidence>(
    m_qxmlStreamReader.attributes().value("id").toString(), pe));
  m_qxmlStreamReader.skipCurrentElement();
  qDebug();
  return true;
}

bool
MzIdentMlReader::readAnalysisCollectionItem()
{

  if(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "SpectrumIdentification")
        {
          readSpectrumIdentification();
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr(
              "Error in MzIdentML/AnalysisCollection, unexpected %1 t")
              .arg(m_qxmlStreamReader.name()));
          m_qxmlStreamReader.skipCurrentElement();
          return false;
        }
      return true;
    }
  return false;
}

void
MzIdentMlReader::readSpectrumIdentification()
{

  m_qxmlStreamReader.skipCurrentElement();
}

bool
MzIdentMlReader::readDataCollectionItem()
{

  if(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "Inputs")
        {
          readInputs();
        }

      // AnalysisData
      else if(m_qxmlStreamReader.name().toString() == "AnalysisData")
        {
          readAnalysisData();
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in MzIdentML/DataCollection, unexpected %1 tag")
              .arg(m_qxmlStreamReader.name()));
          m_qxmlStreamReader.skipCurrentElement();
          return false;
        }
      return true;
    }
  return false;
}

void
MzIdentMlReader::readInputs()
{
  while(m_qxmlStreamReader.readNextStartElement())
    {
      if(m_qxmlStreamReader.name().toString() == "SearchDatabase")
        {
          readSearchDatabase();
        }
      else if(m_qxmlStreamReader.name().toString() == "SpectraData")
        {
          readSpectraData();
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr(
              "Error in MzIdentML/DataCollection/Inputs, unexpected %1 tag")
              .arg(m_qxmlStreamReader.name()));
          m_qxmlStreamReader.skipCurrentElement();
        }
    }
}


//<SearchDatabase numDatabaseSequences="136828"
// location="/home/thierry/test/MS-GF+/Genome_Z_mays_v5a_conta.fasta"
// id="SearchDB_1">
//      <FileFormat>
//        <cvParam cvRef="PSI-MS" accession="MS:1001348" name="FASTA format"/>
//      </FileFormat>
//      <DatabaseName>
//        <userParam name="Genome_Z_mays_v5a_conta.fasta"/>
//      </DatabaseName>
//    </SearchDatabase>
void
MzIdentMlReader::readSearchDatabase()
{

  auto itfasta = m_FastaFileIdMap.find(
    m_qxmlStreamReader.attributes().value("id").toString());

  if(itfasta == m_FastaFileIdMap.end())
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("SearchDatabase id %1 not defined")
          .arg(m_qxmlStreamReader.attributes().value("id")));
    }

  itfasta->second.get()->setFastaSource(
    m_qxmlStreamReader.attributes().value("location").toString());
  m_qxmlStreamReader.skipCurrentElement();
}

void
MzIdentMlReader::readSpectraData()
{

  MsRunSp msrun = mp_project->getMsRunStore().getInstance(
    m_qxmlStreamReader.attributes().value("location").toString());


  // msrun.get()->setXmlId(attributes.value("id"));
  msrun.get()->setSampleName(
    m_qxmlStreamReader.attributes().value("name").toString());


  m_MsRunIdMap.insert(std::pair<QString, MsRunSp>(
    m_qxmlStreamReader.attributes().value("id").toString(), msrun));
  m_qxmlStreamReader.skipCurrentElement();
}

void
MzIdentMlReader::readAnalysisData()
{

  qDebug() << m_qxmlStreamReader.name();
  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();
      if(m_qxmlStreamReader.name().toString() == "SpectrumIdentificationList")
        {
          while(m_qxmlStreamReader.readNextStartElement())
            {
              qDebug() << m_qxmlStreamReader.name();
              if(m_qxmlStreamReader.name().toString() ==
                 "SpectrumIdentificationResult")
                {
                  readSpectrumIdentificationResult();
                }
              else if(m_qxmlStreamReader.name().toString() ==
                      "FragmentationTable")
                {
                  m_qxmlStreamReader.skipCurrentElement();
                }
              else
                {
                  m_qxmlStreamReader.raiseError(
                    QObject::tr("Error in "
                                "MzIdentML/DataCollection/AnalysisData/"
                                "SpectrumIdentificationList, unexpected %1 tag")
                      .arg(m_qxmlStreamReader.name()));
                  m_qxmlStreamReader.skipCurrentElement();
                }
            }
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in MzIdentML/DataCollection/AnalysisData, "
                        "unexpected %1 tag")
              .arg(m_qxmlStreamReader.name()));
          m_qxmlStreamReader.skipCurrentElement();
        }
    }
}

void
MzIdentMlReader::readSpectrumIdentificationResult()
{
  qDebug() << m_qxmlStreamReader.name();
  SpectrumIdentificationResult spectrum_identification_result;
  spectrum_identification_result.cvParamList.clear();
  spectrum_identification_result.userParamList.clear();
  spectrum_identification_result.spectrumIdentificationItemList.clear();

  spectrum_identification_result.spectrumID =
    m_qxmlStreamReader.attributes().value("spectrumID").toString();
  spectrum_identification_result.id =
    m_qxmlStreamReader.attributes().value("id").toString();


  auto itmsrun = m_MsRunIdMap.find(
    m_qxmlStreamReader.attributes().value("spectraData_ref").toString());

  if(itmsrun == m_MsRunIdMap.end())
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("spectraData_ref %1 not defined")
          .arg(m_qxmlStreamReader.attributes()
                 .value("spectraData_ref")
                 .toString()));
    }
  spectrum_identification_result.msrun = itmsrun->second;

  qDebug() << m_qxmlStreamReader.name();

  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();
      if(m_qxmlStreamReader.name() == QString("SpectrumIdentificationItem"))
        {
          qDebug();
          readSpectrumIdentificationItem(spectrum_identification_result);
        }
      else if(m_qxmlStreamReader.name() == QString("cvParam"))
        {
          CvParam cv_param = readCvParam();
          qDebug() << cv_param.toString();
          spectrum_identification_result.cvParamList.push_back(cv_param);
        }

      else if(m_qxmlStreamReader.name() == QString("userParam"))
        {
          UserParam user_param = readUserParam();
          qDebug() << user_param.toString();
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in "
                        "MzIdentML/DataCollection/AnalysisData/"
                        "SpectrumIdentificationList/"
                        "SpectrumIdentificationResult, unexpected %1 tag")
              .arg(m_qxmlStreamReader.name()));
          // m_qxmlStreamReader.skipCurrentElement();
        }
    }

  if(m_qxmlStreamReader.hasError())
    return;
  qDebug() << m_qxmlStreamReader.name();

  // find scan number
  spectrum_identification_result.spectrumIndex = 0;
  spectrum_identification_result.scanNum       = 0;
  spectrum_identification_result.retentionTime = 0;


  //  <cvParam cvRef="PSI-MS" accession="MS:1001115" name="scan number(s)"
  //        value="16079"/>
  for(auto cvParam : spectrum_identification_result.cvParamList)
    {

      qDebug() << cvParam.toString();
      if(cvParam.accession == "MS:1001115")
        {
          spectrum_identification_result.scanNum = cvParam.value.toUInt();
        }
      else if(cvParam.accession == "MS:1003062")
        {
          spectrum_identification_result.spectrumIndex = cvParam.value.toUInt();
        }
      else if((cvParam.accession == "MS:1000016") ||
              (cvParam.accession == "MS:1000894"))
        {
          //[Term]
          // id: MS:1000894
          // name: retention time
          // def: "A time interval from the start of chromatography when an
          // analyte exits a chromatographic column." [PSI:MS]

          //            [Term]
          // id: MS:1000016
          // name: scan start time
          // def: "The time that an analyzer started a scan, relative to the
          // start of the MS run." [PSI:MS]


          spectrum_identification_result.retentionTime =
            cvParam.value.toDouble();
        }
    }
  if((spectrum_identification_result.scanNum == 0) &&
     (spectrum_identification_result.spectrumIndex == 0))
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("scan number or spectrum index not found in "
                    "SpectrumIdentificationResult id %1")
          .arg(spectrum_identification_result.id));
    }


  if(spectrum_identification_result.retentionTime == 0)
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr(
          "retention time not found in SpectrumIdentificationResult id %1")
          .arg(spectrum_identification_result.id));
    }


  for(auto spectrumIdentificationItem :
      spectrum_identification_result.spectrumIdentificationItemList)
    {
      processSpectrumIdentificationItem(spectrum_identification_result,
                                        spectrumIdentificationItem);
    }
}

void
MzIdentMlReader::readSpectrumIdentificationItem(
  MzIdentMlReader::SpectrumIdentificationResult &spectrum_identification_result)
{

  spectrum_identification_result.spectrumIdentificationItemList.push_back(
    SpectrumIdentificationItem());
  spectrum_identification_result.spectrumIdentificationItemList.back()
    .mzidPeptideEvidenceList.clear();
  spectrum_identification_result.spectrumIdentificationItemList.back()
    .cvParamList.clear();
  spectrum_identification_result.spectrumIdentificationItemList.back()
    .userParamList.clear();

  spectrum_identification_result.spectrumIdentificationItemList.back()
    .chargeState =
    m_qxmlStreamReader.attributes().value("chargeState").toUInt();

  spectrum_identification_result.spectrumIdentificationItemList.back()
    .experimentalMassToCharge = m_qxmlStreamReader.attributes()
                                  .value("experimentalMassToCharge")
                                  .toDouble();


  auto itpeptide = m_PeptideIdMap.find(
    m_qxmlStreamReader.attributes().value("peptide_ref").toString());

  if(itpeptide == m_PeptideIdMap.end())
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("peptide_ref %1 not defined")
          .arg(m_qxmlStreamReader.attributes().value("peptide_ref")));
    }
  spectrum_identification_result.spectrumIdentificationItemList.back().peptide =
    itpeptide->second;


  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();
      if(m_qxmlStreamReader.name() == QString("PeptideEvidenceRef"))
        {

          auto itpeptideEvidence =
            m_MzidPeptideEvidenceIdMap.find(m_qxmlStreamReader.attributes()
                                              .value("peptideEvidence_ref")
                                              .toString());

          if(itpeptideEvidence == m_MzidPeptideEvidenceIdMap.end())
            {
              m_qxmlStreamReader.raiseError(
                QObject::tr("peptideEvidence_ref %1 not defined")
                  .arg(m_qxmlStreamReader.attributes().value(
                    "peptideEvidence_ref")));
            }

          spectrum_identification_result.spectrumIdentificationItemList.back()
            .mzidPeptideEvidenceList.push_back(itpeptideEvidence->second);
          m_qxmlStreamReader.skipCurrentElement();
        }
      else if(m_qxmlStreamReader.name() == QString("cvParam"))
        {
          spectrum_identification_result.spectrumIdentificationItemList.back()
            .cvParamList.push_back(readCvParam());
          qDebug() << spectrum_identification_result
                        .spectrumIdentificationItemList.back()
                        .cvParamList.back()
                        .toString();
        }
      else if(m_qxmlStreamReader.name() == QString("userParam"))
        {
          UserParam user_param = readUserParam();
          qDebug() << user_param.toString();
        }
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("Error in "
                        "MzIdentML/DataCollection/AnalysisData/"
                        "SpectrumIdentificationList/"
                        "SpectrumIdentificationResult/"
                        "SpectrumIdentificationItem, unexpected %1 tag")
              .arg(m_qxmlStreamReader.name()));
        }
    }
}


void
MzIdentMlReader::processSpectrumIdentificationItem(
  SpectrumIdentificationResult &spectrum_identification_result,
  const SpectrumIdentificationItem &spectrumIdentificationItem)
{


  if(spectrum_identification_result.msrun.get() != msp_msrun.get())
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr(
          "m_currentSpectrumIdentificationResult.msrun.get() \n%1\n != "
          "msp_msrun.get() \n%2\n")
          .arg(spectrum_identification_result.msrun.get()->toString())
          .arg(msp_msrun.get()->toString()));
    }

  bool isSpectrumIndex = false;
  if(spectrum_identification_result.spectrumIndex != 0)
    {
      spectrum_identification_result.scanNum =
        spectrum_identification_result.spectrumIndex;
      isSpectrumIndex = true;
    }
  PeptideEvidence peptide_evidence(
    msp_msrun.get(), spectrum_identification_result.scanNum, isSpectrumIndex);
  peptide_evidence.setRetentionTime(
    spectrum_identification_result.retentionTime);
  peptide_evidence.setCharge(spectrumIdentificationItem.chargeState);
  peptide_evidence.setPeptideXtpSp(spectrumIdentificationItem.peptide);
  qDebug() << peptide_evidence.getPeptideXtpSp().get()->toAbsoluteString();
  peptide_evidence.setChecked(true);
  peptide_evidence.setIdentificationDataSource(mp_identificationDataSource);
  peptide_evidence.setIdentificationEngine(getIdentificationEngine());


  pappso::pappso_double exp_mass =
    spectrumIdentificationItem.experimentalMassToCharge;
  peptide_evidence.setExperimentalMass(exp_mass);


  //          <cvParam cvRef="PSI-MS" accession="MS:1002049"
  //          name="MS-GF:RawScore" value="356"/> <cvParam cvRef="PSI-MS"
  //          accession="MS:1002050" name="MS-GF:DeNovoScore" value="369"/>
  //          <cvParam cvRef="PSI-MS" accession="MS:1002052"
  //          name="MS-GF:SpecEValue" value="9.149361665076834E-40"/> <cvParam
  //          cvRef="PSI-MS" accession="MS:1002053" name="MS-GF:EValue"
  //          value="2.057944235338586E-32"/>
  //         <userParam name="IsotopeError" value="0"/>
  //         <userParam name="AssumedDissociationMethod" value="HCD"/>
  for(auto cvParam : spectrumIdentificationItem.cvParamList)
    {

      //<cvParam accession="MS:1002258" cvRef="PSI-MS" value="7"
      // name="Comet:matched ions"/>
      if(cvParam.accession == "MS:1002049")
        {
          // PSI-MS MS:1002049 MS-GF:RawScore 356
          peptide_evidence.setParam(PeptideEvidenceParam::msgfplus_raw,
                                    QVariant(cvParam.value.toUInt()));
        }
      else if(cvParam.accession == "MS:1002050")
        {
          // msgfplus_denovo     = 9,  ///< MS:1002050  "MS-GF de novo score."
          peptide_evidence.setParam(PeptideEvidenceParam::msgfplus_denovo,
                                    QVariant(cvParam.value.toUInt()));
        }

      else if(cvParam.accession == "MS:1002052")
        {
          peptide_evidence.setParam(PeptideEvidenceParam::msgfplus_SpecEValue,
                                    QVariant(cvParam.value.toDouble()));
        }

      else if(cvParam.accession == "MS:1002053")
        {
          // PSI-MS MS:1002053 MS-GF:EValue
          peptide_evidence.setParam(PeptideEvidenceParam::msgfplus_EValue,
                                    QVariant(cvParam.value.toDouble()));
        }

      else if(cvParam.accession == "MS:1002054")
        {
          // <cvParam cvRef="PSI-MS" accession="MS:1002054" name="MS-GF:QValue"
          // value="0.0"/>
          peptide_evidence.setParam(PeptideEvidenceParam::msgfplus_QValue,
                                    QVariant(cvParam.value.toDouble()));
        }
      else if(cvParam.accession == "MS:1002055")
        {
          //  <cvParam cvRef="PSI-MS" accession="MS:1002055"
          //  name="MS-GF:PepQValue" value="0.0"/>
          peptide_evidence.setParam(PeptideEvidenceParam::msgfplus_PepQValue,
                                    QVariant(cvParam.value.toDouble()));
        }
      /*
      msgfplus_energy     = 10, ///< MS:1002051  "MS-GF energy score." [PSI:PI]
      msgfplus_SpecEValue = 11, ///< MS:1002052  "MS-GF spectral E-value."
      [PSI:PI] msgfplus_EValue     = 12, ///< MS:1002053  "MS-GF E-value."
      [PSI:PI] msgfplus_isotope_error = 13, ///< MS-GF isotope error
      comet_xcorr   = 14, ///< MS:1002252  "The Comet result 'XCorr'." [PSI:PI]
      comet_deltacn = 15, ///< MS:1002253  "The Comet result 'DeltaCn'."
      [PSI:PI] comet_deltacnstar = 16, ///< MS:1002254  "The Comet result
      'DeltaCnStar'." [PSI:PI] comet_spscore = 17, ///< MS:1002255  "The Comet
      result 'SpScore'." [PSI:PI] comet_sprank  = 18, ///< MS:1002256  "The
      Comet result 'SpRank'." [PSI:PI] comet_expectation_value = 19, ///<
      MS:1002257  "The Comet result 'Expectation value'." [PSI:PI]
        */
      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("cvParam %1 is not taken into account")
              .arg(cvParam.toString()));
        }
    }

  for(auto userParam : spectrumIdentificationItem.userParamList)
    {
      if(userParam.name == "DeepProt:original_count")
        {
          // <userParam name="DeepProt:original_count" value="7"/>
          peptide_evidence.setParam(
            PeptideEvidenceParam::deepprot_original_count,
            QVariant(userParam.value.toUInt()));
        }
      else if(userParam.name == "DeepProt:fitted_count")
        {
          // <userParam name="DeepProt:fitted_count" value="7"/>
          peptide_evidence.setParam(PeptideEvidenceParam::deepprot_fitted_count,
                                    QVariant(userParam.value.toUInt()));
        }
      else if(userParam.name == "DeepProt:match_type")
        {
          peptide_evidence.setParam(
            PeptideEvidenceParam::deepprot_match_type,
            (std::uint8_t)pappso::DeepProtEnumStr::DeepProtMatchTypeFromString(
              userParam.value));
        }
      else if(userParam.name == "DeepProt:status")
        {
          peptide_evidence.setParam(
            PeptideEvidenceParam::deepprot_peptide_candidate_status,
            (std::uint8_t)
              pappso::DeepProtEnumStr::DeepProtPeptideCandidateStatusFromString(
                userParam.value));
        }
      else if(userParam.name == "DeepProt:mass_delta")
        {
          peptide_evidence.setParam(PeptideEvidenceParam::deepprot_mass_delta,
                                    QVariant(userParam.value.toDouble()));
        }
      else if(userParam.name == "DeepProt:delta_positions")
        {
          // DeepProt:delta_positions 4 5 6 7 8 9 10 11 12 13
          peptide_evidence.setParam(
            PeptideEvidenceParam::deepprot_delta_positions, userParam.value);
        }
      // <userParam name="IsotopeError" value="0"/>
      //<userParam name="AssumedDissociationMethod" value="HCD"/>
      /*
    else
      {
        m_errorStr = QObject::tr("userParam %1 is not taken into account")
                       .arg(userParam.toString());
        return false;
      }*/
    }


  for(auto mz_peptide_evidence :
      spectrumIdentificationItem.mzidPeptideEvidenceList)
    {
      PeptideMatch peptide_match;
      peptide_match.setStart(mz_peptide_evidence.start);
      peptide_match.setPeptideEvidenceSp(
        mp_identificationDataSource->getPeptideEvidenceStore().getInstance(
          &peptide_evidence));


      ProteinMatch *p_protein_match =
        mp_identificationGroup->getProteinMatchInstance(
          mz_peptide_evidence.protein.get()->getAccession());

      p_protein_match->setChecked(true);
      // qDebug() << "startElement_protein p_protein_match 3 " <<
      // _p_protein_match;
      p_protein_match->setProteinXtpSp(mz_peptide_evidence.protein);
      p_protein_match->addPeptideMatch(peptide_match);
    }
}
